import requests
import concurrent.futures
import logging
from pathlib import Path
from gwdc_python import GWDC

from .bilby_job import BilbyJob
from .utils import remove_path_anchor, rename_dict_keys

GWCLOUD_ENDPOINT = 'https://gwcloud.org.au/bilby/graphql'
GWCLOUD_FILE_DOWNLOAD_ENDPOINT = 'https://gwcloud.org.au/job/apiv1/file/?fileId='

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
logger.addHandler(ch)


class GWCloud:
    """
    GWCloud class provides an API for interacting with Bilby, allowing jobs to be submitted and acquired.

    Parameters
    ----------
    token : str
        API token for a Bilby user
    endpoint : str, optional
            URL to which we send the queries, by default GWCLOUD_ENDPOINT

    Attributes
    ----------
    client : GWDC
        Handles a lot of the underlying logic surrounding the queries
    """
    def __init__(self, token, endpoint=GWCLOUD_ENDPOINT):
        self.client = GWDC(token=token, endpoint=endpoint)

    def start_bilby_job_from_string(self, job_name, job_description, private, ini_string):
        """Submit the parameters required to start a Bilby job, using the contents of an .ini file

        Parameters
        ----------
        job_name : str
            Name of the job to be created
        job_description : str
            Description of the job to be created
        private : bool
            True if job should be private, False if public
        ini_string : str
            The contents of a Bilby ini file

        Returns
        -------
        str
            Message received from server after job submission
        """
        query = """
            mutation NewBilbyJobFromIniString($input: BilbyJobFromIniStringMutationInput!){
                newBilbyJobFromIniString (input: $input) {
                    result {
                        jobId
                    }
                }
            }
        """

        variables = {
            "input": {
                "params": {
                    "details": {
                        "name": job_name,
                        "description": job_description,
                        "private": private
                    },
                    "iniString": {
                        "iniString": ini_string
                    }
                },
            }
        }

        data = self.client.request(query=query, variables=variables)
        job_id = data['newBilbyJobFromIniString']['result']['jobId']
        return self.get_job_by_id(job_id)

    def start_bilby_job_from_file(self, job_name, job_description, private, ini_file):
        """Submit the parameters required to start a Bilby job, using an .ini file

        Parameters
        ----------
        job_name : str
            Name of the job to be created
        job_description : str
            Description of the job to be created
        private : bool
            True if job should be private, False if public
        ini_file : str or Path
            Path to an .ini file for running a Bilby job

        Returns
        -------
        str
            Message received from server after job submission
        """
        ini_file = Path(ini_file)
        with ini_file.open() as f:
            ini_string = f.read().strip()
            return self.start_bilby_job_from_string(job_name, job_description, private, ini_string)

    def get_preferred_job_list(self):
        """Get list of public Bilby jobs corresponding to a search of "preferred" and a time_range of "Any time"

        Returns
        -------
        list
            List of BilbyJob instances
        """
        return self.get_public_job_list(search="preferred lasky", time_range="Any time")

    def _get_job_model_from_query(self, query_data):
        return BilbyJob(client=self, **rename_dict_keys(query_data, [('id', 'job_id')]))

    def get_public_job_list(self, search="", time_range="Any time", number=100):
        """Obtains a list of public Bilby jobs, filtering based on the search terms
        and the time range within which the job was created.

        Parameters
        ----------
        search : str, optional
            Search terms by which to fileter public job list, by default ""
        time_range : str, optional
            Time range by which to filter job list, by default "Any time"
        number : int, optional
            Number of job results to return in one request, by default 100

        Returns
        -------
        list
            List of BilbyJob instances for the jobs corresponding to the search terms and in the specified time range
        """
        query = """
            query ($search: String, $timeRange: String, $first: Int){
                publicBilbyJobs (search: $search, timeRange: $timeRange, first: $first) {
                    edges {
                        node {
                            id
                            user
                            name
                            description
                        }
                    }
                }
            }
        """

        variables = {
            "search": search,
            "timeRange": time_range,
            "first": number
        }

        data = self.client.request(query=query, variables=variables)

        return [self._get_job_model_from_query(job['node']) for job in data['publicBilbyJobs']['edges']]

    def get_job_by_id(self, job_id):
        """Get a Bilby job instance corresponding to a specific job ID

        Parameters
        ----------
        job_id : str
            ID of job to obtain

        Returns
        -------
        BilbyJob
            BilbyJob instance corresponding to the input ID
        """
        query = """
            query ($id: ID!){
                bilbyJob (id: $id) {
                    id
                    name
                    userId
                    description
                }
            }
        """

        variables = {
            "id": job_id
        }

        data = self.client.request(query=query, variables=variables)

        return self._get_job_model_from_query(data['bilbyJob'])

    def _get_user_jobs(self):
        query = """
            query {
                bilbyJobs {
                    edges {
                        node {
                            id
                            name
                            userId
                            description
                        }
                    }
                }
            }
        """
        
        data = self.client.request(query=query)

        return [self._get_job_model_from_query(job['node']) for job in data['bilbyJobs']['edges']]

    def _get_files_by_job_id(self, job_id):
        query = """
            query ($jobId: ID!) {
                bilbyResultFiles (jobId: $jobId) {
                    files {
                        path
                        isDir
                        fileSize
                        downloadToken
                    }
                }
            }
        """

        variables = {
            "jobId": job_id
        }

        data = self.client.request(query=query, variables=variables)

        file_list = []
        for f in data['bilbyResultFiles']['files']:
            if f['isDir']:
                continue
            f.pop('isDir')
            f['path'] = remove_path_anchor(Path(f['path']))
            file_list.append(f)

        return file_list

    def _get_file_by_id(self, file_id):
        download_url = GWCLOUD_FILE_DOWNLOAD_ENDPOINT + str(file_id)
        request = requests.get(download_url)
        return request.content

    def _get_files_by_id(self, file_ids):
        result = []
        with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
            files = executor.map(self._get_file_by_id, file_ids)
            for i, f in enumerate(files):
                logger.info(f'File {i+1} of {len(file_ids)} downloaded! Filesize: {len(f)}')
                result.append(f)

        return result

    def _get_download_id_from_token(self, job_id, file_token):
        """Get a single file download id for a file download token

        Parameters
        ----------
        job_id : str
            Job id which owns the file token

        file_token : str
            Download token for the desired file

        Returns
        -------
        str
            Download id for the desired file
        """
        return self._get_download_ids_from_tokens(job_id, [file_token])[0]

    def _get_download_ids_from_tokens(self, job_id, file_tokens):
        """Get many file download ids for a list of file download tokens

        Parameters
        ----------
        job_id : str
            Job id which owns the file token

        file_tokens : list
            Download tokens for the desired files

        Returns
        -------
        list
            List of download ids for the desired files
        """
        query = """
            mutation ResultFileMutation($input: GenerateFileDownloadIdsInput!) {
                generateFileDownloadIds(input: $input) {
                    result
                }
            }
        """

        variables = {
            "input": {
                "jobId": job_id,
                "downloadTokens": file_tokens
            }
        }

        data = self.client.request(query=query, variables=variables)

        return data['generateFileDownloadIds']['result']



from pathlib import Path
from gwcloud_python.utils import identifiers


def test_file_suffix_funcs():
    png_test_path = Path('/this/is/a/test.png')
    html_test_path = Path('/this/is/a/test.html')
    no_suffix_test_path = Path('/this/is/a/test')

    assert identifiers.png_file(png_test_path) is True
    assert identifiers.png_file(html_test_path) is False
    assert identifiers.png_file(no_suffix_test_path) is False

    assert identifiers.html_file(png_test_path) is False
    assert identifiers.html_file(html_test_path) is True
    assert identifiers.png_file(no_suffix_test_path) is False


def test_file_beginning_funcs():
    data_dir_test_path = Path('/data/this/is/a/test')
    relative_data_dir_test_path = Path('data/this/is/a/test')
    result_dir_test_path = Path('/result/this/is/a/test')
    relative_result_dir_test_path = Path('result/this/is/a/test')
    no_dir_test_path1 = Path('data.png')
    no_dir_test_path2 = Path('/data.png')

    assert identifiers.data_dir(data_dir_test_path) is False
    assert identifiers.data_dir(relative_data_dir_test_path) is True
    assert identifiers.data_dir(result_dir_test_path) is False
    assert identifiers.data_dir(relative_result_dir_test_path) is False
    assert identifiers.result_dir(data_dir_test_path) is False
    assert identifiers.result_dir(relative_data_dir_test_path) is False
    assert identifiers.result_dir(result_dir_test_path) is False
    assert identifiers.result_dir(relative_result_dir_test_path) is True
    assert identifiers.data_dir(no_dir_test_path1) is False
    assert identifiers.data_dir(no_dir_test_path2) is False


def test_file_ending_funcs():
    config_complete_test_path = Path('/this/is/a_config_complete.ini')
    merged_json_test_path = Path('/this/is/a_merge_result.json')
    corner_plot_test_path = Path('/this/is/a_corner.png')
    no_suffix_test_path = Path('/this/is/a/test')

    assert identifiers.config_file(config_complete_test_path) is True
    assert identifiers.config_file(merged_json_test_path) is False
    assert identifiers.config_file(corner_plot_test_path) is False
    assert identifiers.config_file(no_suffix_test_path) is False

    assert identifiers.merged_json_file(config_complete_test_path) is False
    assert identifiers.merged_json_file(merged_json_test_path) is True
    assert identifiers.merged_json_file(corner_plot_test_path) is False
    assert identifiers.merged_json_file(no_suffix_test_path) is False

    assert identifiers.corner_plot_file(config_complete_test_path) is False
    assert identifiers.corner_plot_file(merged_json_test_path) is False
    assert identifiers.corner_plot_file(corner_plot_test_path) is True
    assert identifiers.corner_plot_file(no_suffix_test_path) is False



from dataclasses import dataclass, field
import logging
from .utils import identifiers, write_file_at_path

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
logger.addHandler(ch)


@dataclass(init=False)
class BilbyJob:
    """
    BilbyJob class is useful for interacting with the Bilby jobs returned from a call to the GWCloud API.
    It is primarily used to store job information and obtain files related to the job.

    Parameters
    ----------
    client : ~gwcloud_python.gwcloud.GWCloud
        A reference to the GWCloud object instance from which the BilbyJob was created
    job_id : str
        The id of the Bilby job, required to obtain the files associated with it
    name : str
        Job name
    description : str
        Job description
    kwargs : dict, optional
        Extra arguments, stored in `other` attribute
    """
    client: object = field(repr=False)
    job_id: str
    name: str
    description: str
    other: dict  #: Used to retain reference to other arbitrary information returned from the GWCloud query

    def __init__(self, client, job_id, name, description, **kwargs):
        self.client = client
        self.job_id = job_id
        self.name = name
        self.description = description
        self.other = kwargs

    def _get_files_from_file_list(self, file_list):
        file_ids, file_paths = [], []

        for f in file_list:
            file_ids.append(f['downloadToken'])
            file_paths.append(f['path'])

        files = self.get_files_by_tokens(file_ids)

        return list(zip(file_paths, files))

    def _save_files(self, root_path, files, preserve_directory_structure):
        for i, (file_path, file_contents) in enumerate(files):
            write_file_at_path(root_path, file_path, file_contents, preserve_directory_structure)
            logger.info(f'File {i+1} of {len(files)} saved : {file_path}')

        return 'Files saved!'

    def _get_file_list_subset(self, identifier):
        file_list = self.get_full_file_list()
        return [f for f in file_list if identifier(f['path'])]

    def get_full_file_list(self):
        """Get information for all files associated with this job

        Returns
        -------
        list
            List of dicts containing information on the files
        """
        return self.client._get_files_by_job_id(self.job_id)

    def get_file_by_token(self, file_token):
        """Get the contents of a file

        Parameters
        ----------
        file_token : str
            Download token for the desired file

        Returns
        -------
        bytes
            Content of the file
        """
        download_id = self.client._get_download_id_from_token(self.job_id, file_token)

        return self.client._get_file_by_id(download_id)

    def get_files_by_tokens(self, file_tokens):
        """Get the contents of files

        Parameters
        ----------
        file_tokens : list
            List of download tokens

        Returns
        -------
        list
            Contents of the files
        """
        download_ids = self.client._get_download_ids_from_tokens(self.job_id, file_tokens)

        return self.client._get_files_by_id(download_ids)

    def get_default_file_list(self):
        """Get information for the default files associated with this job

        Returns
        -------
        list
            List of dicts containing information on the files
        """
        return self._get_file_list_subset(identifiers.default_file)

    def get_config_file_list(self):
        """Get information for the config files associated with this job

        Returns
        -------
        list
            List of dicts containing information on the files
        """
        return self._get_file_list_subset(identifiers.config_file)

    def get_png_file_list(self):
        """Get information for the PNG files associated with this job

        Returns
        -------
        list
            List of dicts containing information on the files
        """
        return self._get_file_list_subset(identifiers.png_file)

    def get_corner_plot_file_list(self):
        """Get information for the PNG files associated with this job

        Returns
        -------
        list
            List of dicts containing information on the files
        """
        return self._get_file_list_subset(identifiers.corner_plot_file)

    def get_default_files(self):
        """Obtain the content of all the default files

        Returns
        -------
        list
            List containing tuples of the file path and associated file contents
        """
        file_list = self.get_default_file_list()
        return self._get_files_from_file_list(file_list)

    def save_default_files(self, root_path, preserve_directory_structure=True):
        """Save the default files

        Parameters
        ----------
        root_path : str or pathlib.Path
            The base directory
        preserve_directory_structure : bool, optional
            Save the files in the same structure that they were downloaded in, by default True
        """
        files = self.get_default_files()
        self._save_files(root_path, files, preserve_directory_structure)

    def get_config_files(self):
        """Obtain the content of all the config files

        Returns
        -------
        list
            List containing tuples of the file path and associated file contents
        """
        file_list = self.get_config_file_list()
        return self._get_files_from_file_list(file_list)

    def save_config_files(self, root_path, preserve_directory_structure=True):
        """Save the config files

        Parameters
        ----------
        root_path : str or pathlib.PPath
            The base directory
        preserve_directory_structure : bool, optional
            Save the files in the same structure that they were downloaded in, by default True
        """
        files = self.get_config_files()
        self._save_files(root_path, files, preserve_directory_structure)

    def get_png_files(self):
        """Obtain the content of all the PNG files

        Returns
        -------
        list
            List containing tuples of the file path and associated file contents
        """
        file_list = self.get_png_file_list()
        return self._get_files_from_file_list(file_list)

    def save_png_files(self, root_path, preserve_directory_structure=True):
        """Save the PNG files

        Parameters
        ----------
        root_path : str or pathlib.Path
            The base directory
        preserve_directory_structure : bool, optional
            Save the files in the same structure that they were downloaded in, by default True
        """
        files = self.get_png_files()
        self._save_files(root_path, files, preserve_directory_structure)

    def get_corner_plot_files(self):
        """Obtain the content of all the corner plot files

        Returns
        -------
        list
            List containing tuples of the file path and associated file contents
        """
        file_list = self.get_corner_plot_file_list()
        return self._get_files_from_file_list(file_list)

    def save_corner_plot_files(self, root_path, preserve_directory_structure=True):
        """Save the corner plot files

        Parameters
        ----------
        root_path : str or pathlib.Path
            The base directory
        preserve_directory_structure : bool, optional
            Save the files in the same structure that they were downloaded in, by default True
        """
        files = self.get_corner_plot_files()
        self._save_files(root_path, files, preserve_directory_structure)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gwcloud_python',
 'gwcloud_python.tests',
 'gwcloud_python.utils',
 'gwcloud_python.utils.tests']

package_data = \
{'': ['*'], 'gwcloud_python': ['.pytest_cache/*', '.pytest_cache/v/cache/*']}

install_requires = \
['graphene-file-upload>=1.3.0,<2.0.0',
 'gwdc-python>=0.1.0,<0.2.0',
 'importlib-metadata>=4.5.0,<5.0.0',
 'jwt>=1.2.0,<2.0.0',
 'requests>=2.25.1,<3.0.0']

extras_require = \
{'docs': ['Sphinx>=4.0.2,<5.0.0', 'sphinx-rtd-theme>=0.5.2,<0.6.0']}

setup_kwargs = {
    'name': 'gwcloud-python',
    'version': '0.3.0',
    'description': 'Wrapper of GWDC API, used for interacting with the GWCloud endpoints',
    'long_description': 'GWCloud Python API\n==================\n\nGWCloud is a service used to handle both the submission of `Bilby <https://pypi.org/project/bilby/>`_ jobs to a supercomputer queue and the obtaining of the results produced by these jobs.\nWhile there is a web interface for this service, which is recommended for beginners, this package can be used to allow Bilby job submission and manipulation from Python scripts.',
    'author': 'Thomas Reichardt',
    'author_email': 'treichardt@swin.edu.au',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/gravitationalwavedc/gwcloud_python',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
